#
#    Copyright (C) 2003-2005  James Tauber
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#


__provider__      = "PageProvider"
__contributions__ = {}


import re
import mimetypes

from leonardo.formatters.manager import FormatterManager
from leonardo.web_resource import PagePart, TopLevelResource, FileObjectResource

import page_templates



class PageProvider:
    
    
    def __init__(self, resource_manager, lfs, config):
    
        self.resource_manager  = resource_manager
        self.lfs               = lfs
        self.config            = config

        self.home_key          = config.get("page", "home_key")

        self.formatter_manager = FormatterManager(config)


    def get(self, resource_id, request, main_resource):

        page_part = self.get_page(resource_id, request, main_resource)
            
        if main_resource: # if there's a parent resource, we're done
            return page_part

        if isinstance(page_part, NonWikiPage): # if it's a non wiki page, we're done
            return page_part

        key = page_part.get_id()

        if page_part.exists():
            return HTMLPage(key, page_part, self.resource_manager, request, main_resource, self.config)
                
        else: # new page
            if request.logged_in():
                return PageDoesNotExist(key, self.resource_manager, request, main_resource, self.config)
            else:
                return HTMLPage404(key, self.resource_manager, request, main_resource, self.config)


    def post(self, resource_id, request, main_resource):

        page_part = self.get_page(resource_id, request, main_resource)

        for provider in self.resource_manager.get_contributors("post_processing"):
            result = provider(request, page_part, None, self.config)
            if result:
                return result
            
        # @@@ what can we do about duplication here

        if isinstance(page_part, NonWikiPage): # if it's a non wiki page, we're done
            return page_part

        key = page_part.get_id()

        if page_part.exists():
            return HTMLPage(key, page_part, self.resource_manager, request, main_resource, self.config)
                
        else: # new page
            if request.logged_in():
                return PageDoesNotExist(key, self.resource_manager, request, main_resource, self.config)
            else:
                return HTMLPage404(key, self.resource_manager, request, main_resource, self.config)



    def get_page(self, resource_id, request, main_resource):
        if resource_id == "":
            key = self.home_key
        else:
            key = resource_id

        file_object = self.lfs.get(key)
        content_type = file_object.get_content_type()
        
        if file_object.exists() and not self.formatter_manager.known(content_type):
            mime_type, encoding = mimetypes.guess_type("dummy.%s" % content_type)
            if not mime_type:
                mime_type = "application/octet-stream"
            return NonWikiPage(key, file_object, mime_type)
    
        return WikiPagePart(key, self.resource_manager, request, main_resource, self.config, file_object, self.formatter_manager)



class NonWikiPage(FileObjectResource, TopLevelResource):

    
    def __init__(self, resource_id, file_object, mime_type):
        TopLevelResource.__init__(self, resource_id, mime_type)
        FileObjectResource.__init__(self, file_object)
        
        
    def is_read_only(self):
        return False
        


class HTMLPage(TopLevelResource):


    def __init__(self, resource_id, page_part, resource_manager, request, main_resource, config):
        TopLevelResource.__init__(self, resource_id, "text/html; charset=utf-8")

        self.resource_manager = resource_manager
        self.request          = request
        self.main_resource    = main_resource
        self.config           = config

        self.home_href         = config.get("DEFAULT", "home_href")
        self.site_title        = config.get("page", "site_title")
        self.site_sub_title    = config.get("page", "site_sub_title")
        self.rights            = config.get("page", "rights")
        self.home_key          = config.get("page", "home_key")
        self.page_template_key = config.get("page", "page_template_key")
        self.menu_key          = config.get("menu", "menu_key")

        self.make_html(page_part, request)


    def exists(self):
        return True
                
        
    def get_content(self):
        return self.html    


    def is_read_only(self):
        return True


    def make_html(self, page_part, request):
        page_title = page_part.get_title()
        if page_title:
            title = "%s : %s" % (self.site_title, page_title)
        else:
            title = self.site_title
        content = page_part.get_html()
        lastmod = page_part.get_lastmod()
        author_email = page_part.get_property("author")
        if author_email:
            author_name = request.get_name(author_email)
        else:
            author_name = None
        menu = self.resource_manager.get(self.menu_key, request, page_part).get_html()
        self.html = self.page_template(title, page_title, menu, content, request, page_part, lastmod, author_name)
        

    def page_template(self, title, main_title, menu, content, request, page_part, lastmod=None, author=None):
        if lastmod:
            import time
            lastmod = time.strftime(page_templates.time_format, time.gmtime(lastmod))
            lastmod_display = page_templates.lastmod_display % locals()
        else:
            lastmod_display = ""
            
        if self.site_sub_title:
            site_sub_title = self.site_sub_title
            sub_title      = page_templates.sub_title % locals()
        else:
            sub_title = ""

        if author:
            author_display = "by %s" % author
        else:
            author_display = ""

        site_title       = self.site_title
        home_page        = self.home_href
        rights           = self.rights

        head_links = ""
        for provider in self.resource_manager.get_contributors("head_links"):
            head_links += provider(self.request, self, self.main_resource, self.config)

        if main_title:
            title_html = page_templates.title_template % locals()
        else:
            title_html = ""

        page_template = self.resource_manager.get(self.page_template_key, request, "dummy")

        return page_template.get_html() % locals()



class ForbiddenPage(HTMLPage):


    def __init__(self, resource_id, resource_manager, request, main_resource, config):
        HTMLPage.__init__(self, resource_id, None, resource_manager, request, main_resource, config)
        self.status = "403 Forbidden"
        

    def make_html(self, page_part, request):
    
        site_title = self.site_title
        title = page_templates.operation_not_allowed_title % locals()
        content = page_templates.operation_not_allowed % locals()
        time = None
        menu = self.resource_manager.get(self.menu_key, request, self).get_html()
        self.html = self.page_template(title, title, menu, content, request, page_part, time)
        


class PageDoesNotExist(HTMLPage):


    def __init__(self, resource_id, resource_manager, request, main_resource, config):
        HTMLPage.__init__(self, resource_id, None, resource_manager, request, main_resource, config)
        self.status = "404 Not Found"


    def make_html(self, page_part, request):
    
        site_title = self.site_title
        title = page_templates.page_does_not_exist_title % locals()
        content = page_templates.page_does_not_exist % locals()
        time = None
        menu = self.resource_manager.get(self.menu_key, request, self).get_html()
        self.html = self.page_template(title, title, menu, content, request, page_part, time)
        
        
    def is_read_only(self):
        return False



class HTMLPage404(HTMLPage):


    def __init__(self, resource_id, resource_manager, request, main_resource, config):
        HTMLPage.__init__(self, resource_id, None, resource_manager, request, main_resource, config)
        self.status = "404 Not Found"


    def make_html(self, page_part, request):
    
        site_title = self.site_title
        title = page_templates.title_404 % locals()
        content = page_templates.content_404 % locals()
        time = None
        menu = self.resource_manager.get(self.menu_key, request, self).get_html()
        self.html = self.page_template(title, title, menu, content, request, page_part, time)
        
        
    def is_read_only(self):
        return True



class WikiPagePart(FileObjectResource, PagePart):

    # @@@ perhaps things like resource_manager, request, config should be moved to superclass __init__
    def __init__(self, resource_id, resource_manager, request, main_resource, config, file_object, formatter_manager):

        PagePart.__init__(self, resource_id)
        FileObjectResource.__init__(self, file_object)

        self.resource_manager  = resource_manager
        self.request           = request
        self.main_resource     = main_resource
        self.config            = config
        self.formatter_manager = formatter_manager

        self.title = None


    def get_html(self):
        content = self.get_content()
        content_type = self.get_content_type()
            
        html = self.formatter_manager.format(content, content_type, self.file_object, self.resource_getter_)
        for provider in self.resource_manager.get_contributors("page_end"):
            html += provider(self.request, self, self.main_resource, self.config)
        return html
        

    def resource_getter_(self, resource_id):
        return self.resource_manager.get(resource_id, self.request, self)
    

    def get_title(self):
        return self.get_property("page_title")
        
        
    def is_read_only(self):
        return False
        